{*********************************************************}
{                                                         }
{    Calmira Visual Component Library 1.0                 }
{    by Li-Hsin Huang,                                    }
{    released into the public domain January 1997         }
{                                                         }
{---------------------------------------------------------}
{ all the singe Vcl unit are moved in one big vcl unit    }
{ by Bauer Karl e-mail: karbauer@mail.teleconsult.de      }
{---------------------------------------------------------}
{ If it breaks your data, PC, reputation                  }
{or anything else you have my condolences. But you don't  }
{get a guarantee, so I can't acceept any responsibility...}
{*********************************************************}
{ TCalAppHolder component }

{ TCalAppHolder is a simple container component that provides design-time
  access to TApplication's properties and events.  When the component
  is loaded, it assigns its data to TApplication.

  You can use it to automatically create event handlers for TApplication
  and TScreen just like events for visual controls -- just double click
  on the event.

  In addition, there is the OnWndProc event which is, in effect, a fast
  way to subclass the application's main window without going down to the
  API level.  If you create an OnWndProc event, this is triggered each
  time the application's WndProc is called, and you can process messages
  that never generate an OnMessage event because they are sent directly
  to the window procedure.

  For this event, return False to allow TApplication to continue processing
  the message, and True to stop TApplication handling it further.  It is
  especially useful to trap WM_ENDSESSION here so that your program can
  save its layout etc. before Windows shuts down.
}
{ TCalCheckList control }

{ TCalCheckList is a listbox that acts as an array of checkboxes.  It
  draws each item like a 3D check box, using the Selected property
  to determine if an item is checked.  At design time, use the Items
  property to set the contents.

  This has many advantages over a large array or group of standard
  TCheckbox controls:

  1.  Less resources -- a listbox only uses one window handle.
  2.  Fast -- only one control is created and drawn.
  3.  Practically unlimited capacity, without using more resources.
  4.  Scrollable -- pack a large number of options into a small
      space in a dialog box.

  The check list was originally designed so that many Boolean variables
  can be set without the need to map each var to one TCheckbox, which is
  error prone and slow, so there are two additional methods to allow
  efficient data transfer.

  SetData - accepts an open array of Booleans which is used to set
    the Selected property.
  GetData - accepts an open array of Boolean pointers, which is
    assigned the values from the Selected property.
  RangeCheck - if True, a call to GetData or SetData will check that
    the size of the open array matches the size of the list.  This
    often catches out ommissions and inconsistencies.

  For long lists, you can use just paste a copy the SetData call,
  change the "Set" to a "Get" and add @ symbols in front of each
  boolean identifier.

  Example :

  var
    DebugInfo, LocalSymbols, SymbolInfo : Boolean;

  CheckList1.SetData([DebugInfo, LocalSymbols, SymbolInfo]);
  if ShowModal = mrOK then
  CheckList1.GetData([@DebugInfo, @LocalSymbols, @SymbolInfo]);

  Don't forget to distribute CHKLIST.RES, which contains the
  fake checkbox bitmaps.
}
{ TCalMultiGrid component

  Properties

  Selected - determines if a given cell is highlighted
  SelCount - determines how many cells are highlighted in the grid
  Multi - true if the grid is in "multi-select" mode
  AllowMulti - enables or disables multiple selections
  Limit - the valid range for the grid.  Cells with an index outside
    of Limit will not be painted and cannot be selected with the mouse.
  Focus - determines which cell has the dotted box draw around it
  ThumbTrack - controls the goThumbTracking element of TCustomGrid
    (the inherited Options property is not made public)
  DropFocus - determines which cell has a focus rect drawn around it
    during drag and drop.  Set to -1 to hide the drop focus.

  Methods

  SelectAll and DeselectAll - highlights and unhighlights all cells in
    the grid, up to Limit
  CellIndex - returns the linear index of a given row and column
  Reset - deselects all cells without generating events and redraws
    the control Use this to initialize between different phases of use.
  Select - moves the focus to the given cell and selects it
  MouseToCell - returns the index of the cell at the given pixel position
  SetSize - changes the number of columns and rows while preserving the
    current selection.  If you modify the ColCount and RowCount
    properties directly, all selections are lost.
  SizeGrid - automatically adjusts the number of columns and rows to
    fit the current grid size

  Events

  OnSelectCell - occurs just before a cell is selected (like TDrawGrid's
     OnSelectCell event).  You have the chance to cancel this operation.

  OnSelect - occurs after the user has selected a cell by left clicking
     with the mouse (or moving the cursor keys).  Typically you would
     use this event to respond to a single or multiple selection.  This
     event occurs only once for each mouse click.

  OnCellSelected - occurs after the highlight of a cell is turned on or
     off, either by the user or by the program assigning a value to the
     Selected property.  If the user selects a range of cells by using
     the Shift key, this event occurs once for every cell that has its
     highlight changed.

  OnDrawCell - same as OnDrawCell for a TDrawGrid except that an integer
     cell index is used
}
{ TCalDropClient component

  Allows Delphi programs to accept file drops from File Manager.
  Another ubiquitous component, but I couldn't find a freeware one
  with source code so I wrote my own.  Since I already had a handler
  for Application.OnMessage, I decided against low level subclassing.
  Consequently, you'll need to call TCalDropClient.CheckMessage inside
  the OnMessage handler.

  Files : TStrings; (inherited, run-time only)
    Contains a list of dropped files after another program has
    dropped them.  This list is cleared before each drop if the
    AutoClear property is True.

  AutoClear : Boolean (inherited)
    Determines whether to clear the Files list before a drop.

  Handle : HWND (read and run-time only)
    The window handle of the owning windowed control

  DropPos : TPoint (read and run-time only)
    The coordinates of the last drop, in the coordinate system
    of the owning windowed control (usually a TForm)

  OnDropFiles
    When this event occurs, the Files property contains a list of
    dropped files and DropPos contains the coordinates of the cursor
    when the drop occured.

  CheckMessage(var Msg : TMsg; var Handled : Boolean);
    You need to call this message in the application's OnMessage handler.
    It checks an application message record to see if it is a file drop
    message.  If so, the corresponding DropClient component is activated
    to trigger the OnDropFiles event which you can handle.
    Handled is set to True if the message has been processed.
}
{ TDragDrop component }

{ TDragDrop is an abstract base class for TCalDropServer and TCalDropClient.
  It defines the Files and AutoClear properties.  To add more
  funtionality to both TCalDropServer or TCalDropClient (e.g., methods to
  process filenames), write the code for TDragDrop instead.
}

{ TCalDropServer component

  If you ever need to drop files into other programs, this component
  can lend a hand, although it still requires some programming on your
  part.  In a control's OnMouseMove handler, call CanDrop to determine
  if if the cursor is over a suitable window.  In the OnEndDrag handler,
  call the DragFinished method.  If a drop is allowed, the OnFileDrop
  event is triggered which lets you assign the files to drop.

  AutoClear : Boolean (inherited)
    If this is True, then the Files property is cleared after each
    drop is completed.

  Files : TStrings (run-time only)
    Contains a list of files to drop into another program

  InternalDrop : Boolean
    If this is set to True, OnFileDrag events will occur when the
    cursor is over a valid window belonging to your program.  If it is
    False, OnFileDrag only occurs when the cursor is over another program.
    Generally, you should use Delphi's own drag and drop handling and set
    this to False.

  DesktopDrop : Boolean
    If this is set to True, you will receive an OnDesktopDrop event when
    the mouse is released over the desktop background or wallpaper.

  CanDrop: Boolean;
     Returns True if the cursor is over something that can accept drops.
     Call this inside an OnMouseMove handler.  If the cursor is currently
     over a window, the call will trigger the OnFileDrag event to ask you
     for confirmation of the drop. 

  procedure DragFinished;
     Call this inside an OnEndDrag handler.  If the cursor is over a
     suitable window and you have responded to the OnFileDrag event,
     then an OnFileDrop is triggered to let you assign the filenames.

  procedure DropFiles(Wnd: HWnd; AMousePos: TPoint);
     Encapsulates the WM_DROPFILES message and immediately causes a
     drop into the given window.  The strings in the Files property are
     contatenated into the required structure and a WM_DROPFILES message
     is sent to the given window.  The TPoint parameter lets you control
     the location of this forced drop.

  OnFileDrag
     Occurs when you call CanDrop and the cursor is over a window that
     accepts dropped files.  To permit the drop (if the user releases the
     mouse button), set the Accept property to True.  The Target parameter
     contains the handle of the window in question, and lets you perform
     your own tests before accepting.

  OnFileDrop
     Occurs when the user releases the mouse button over a window that
     accepts files and you call DragFinished method.  During this event,
     you must fill the Files property with the files you wish to drop
     (one file per line).  The drop takes place as soon as your handler
     finishes executing.

  OnDesktopDrop
     Occurs when the user releases the mouse button over the desktop.
     This is only a notification - TCalDropServer doesn't do anything
     afterwards.  The Target parameter contains the window handle of
     the desktop, in case you want to send any messages to it.
}


{ TCalSystemMenu provides a thin wrapper for the Windows API menu functions,
  and is used to change a form's system menu.  It's most useful in the
  OnCreate handler, when you can modify the menu before the form appears.
  Mainly, it saves you having to remember (or look up) the multitude of
  parameters. }

{ TCalFormDrag component }

{ TCalFormDrag lets the user move and resize a form with a non-resizable
  border, typically bsSingle.  It can simulate a "solid" drag like
  Windows95, or for slower machines, it can use hollow boxes instead.
  Just drop the component onto your form, and it is ready to use.

  Windows 3.x is generally not up to dynamically resizing the controls
  during dragging, so TCalFormDrag hides all the controls and shows them
  when the dragging has stopped.

  Note that you should not re-assign the following event handlers
  of TForm at run-time: OnMouseMove, OnMouseDown, OnMouseUp,
  OnPaint and OnResize.

  TCalFormDrag prevents some of these events from occuring when it
  takes over the dragging, so that you need not be concerned with
  any side effects -- just pretend the component is not there and
  program normally.

  DragWidth
    The size of the square in the bottom right corner reserved for
    resizing the form.

  Hollow
    Determines if the form is resized during dragging, or a hollow
    rectangle used to represent the form's dimensions.

  MinWidth, MinHeight, MaxWidth, MaxHeight
    Constrains the size of the form

  AllowMove, AllowSize
    Enables the two operations of this component

  DragState (run-time and read only)
    Indicates what kind of dragging is taking place.  Check this to
    avoid executing code that could disrupt the operation.
}

{ Some useful Delphi and Windows routines }

unit Calmuniv;

interface

uses
  SysUtils, WinTypes, WinProcs, Messages, Classes, Graphics, Controls,
  Forms, Dialogs, StdCtrls, Buttons, Grids, Menus;


type
  TIconSelForm = class(TForm)
    Grid: TDrawGrid;
    CancelBtn: TBitBtn;
    Label1: TLabel;
    Label2: TLabel;
    OpenDialog: TOpenDialog;
    OKBtn: TBitBtn;
    FileEdit: TEdit;
    BrowseBtn: TSpeedButton;
    procedure FormCreate(Sender: TObject);
    procedure FormDestroy(Sender: TObject);
    procedure GridDrawCell(Sender: TObject; Col, Row: Longint; Rect: TRect;
      State: TGridDrawState);
    procedure GridSelectCell(Sender: TObject; Col, Row: Longint;
      var CanSelect: Boolean);
    procedure FormShow(Sender: TObject);
    procedure BrowseBtnClick(Sender: TObject);
  private
    { Private declarations }
    Icons : TList;
    FIndex : Integer;
    procedure SearchFile(const filename: TFilename);
  public
    { Public declarations }
    property Index : Integer read FIndex write FIndex;
  end;

type
  TCalIconDialog = class(TComponent)
  private
    { Private declarations }
    FFilename  : TFilename;
    FIconIndex : Word;
    FHistory : TStrings;
    procedure SetHistory(value : TStrings);
  protected
    { Protected declarations }
  public
    { Public declarations }
    constructor Create(AOwner : TComponent); override;
    destructor Destroy; override;
    function Execute : Boolean;
  published
    { Published declarations }
    property Filename : TFilename read FFilename write FFilename;
    property IconIndex : Word read FIconIndex write FIconIndex default 0;
    property HistoryList : TStrings read FHistory write SetHistory;
  end;


type
  TCalAppHolder = class(TComponent)
  private
    { Private declarations }
    FHelpFile : string;
    FHintColor : TColor;
    FHintPause : Integer;
    FTitle : string;
    FOnActivate : TNotifyEvent;
    FOnDeactivate : TNotifyEvent;
    FOnException : TExceptionEvent;
    FOnHelp : THelpEvent;
    FOnHint : TNotifyEvent;
    FOnIdle : TIdleEvent;
    FOnMessage : TMessageEvent;
    FOnMinimize : TNotifyEvent;
    FOnRestore : TNotifyEvent;
    FOnShowHint : TShowHintEvent;
    FOnWndProc : TWindowHook;
    FOnActiveControlChange : TNotifyEvent;
    FOnActiveFormChange : TNotifyEvent;
  protected
    { Protected declarations }
    procedure Loaded; override;
  public
    { Public declarations }
    constructor Create(AOwner : TComponent); override;
    destructor Destroy; override;
  published
    { Published declarations }
    property HelpFile : string read FHelpfile write FHelpFile;
    property HintColor : TColor read FHintColor write FHintColor;
    property HintPause : Integer read FHintPause write FHintPause default 800;
    property Title : string read FTitle write FTitle;
    property OnActivate : TNotifyEvent read FOnActivate write FOnActivate;
    property OnDeactivate : TNotifyEvent read FOnDeactivate write FOnDeactivate;
    property OnException : TExceptionEvent read FOnException write FOnException;
    property OnHelp : THelpEvent read FOnHelp write FOnHelp;
    property OnHint : TNotifyEvent read FOnHint write FOnHint;
    property OnIdle : TIdleEvent read FOnIdle write FOnIdle;
    property OnMessage : TMessageEvent read FOnMessage write FOnMessage;
    property OnMinimize : TNotifyEvent read FOnMinimize write FOnMinimize;
    property OnRestore : TNotifyEvent read FOnRestore write FOnRestore;
    property OnShowHint : TShowHintEvent read FOnShowHint write FOnShowHint;
    property OnWndProc : TWindowHook read FOnWndProc write FOnWndProc;
    property OnActiveControlChange : TNotifyEvent read
      FOnActiveControlChange write FOnActiveControlChange;
    property OnActiveFormChange : TNotifyEvent read
      FOnActiveFormChange write FOnActiveFormChange;
  end;

type
  TBarKind = (bkHorizontal, bkVertical);

  TCalBarGauge = class(TGraphicControl)
  private
    { Private declarations }
    FMinValue: Longint;
    FMaxValue: Longint;
    FCurValue: Longint;
    FShowText: Boolean;
    FBorderStyle: TBorderStyle;
    FForeColor: TColor;
    FBackColor: TColor;
    FCtl3D : Boolean;
    FKind : TBarKind;
    procedure SetShowText(Value: Boolean);
    procedure SetBorderStyle(Value: TBorderStyle);
    procedure SetForeColor(Value: TColor);
    procedure SetBackColor(Value: TColor);
    procedure SetMinValue(Value: Longint);
    procedure SetMaxValue(Value: Longint);
    procedure SetProgress(Value: Longint);
    procedure SetCtl3D(Value: Boolean);
    procedure SetKind(Value: TBarKind);
    function GetPercentDone: Integer;
  protected
    { Protected declarations }
    procedure Paint; override;
  public
    { Public declarations }
    constructor Create(AOwner: TComponent); override;
    procedure AddProgress(Value: Longint);
    property PercentDone: Integer read GetPercentDone;
  published
    { Published declarations }
    property Align;
    property Ctl3D : Boolean read FCtl3D write SetCtl3D default True;
    property Enabled;
    property ShowText: Boolean read FShowText write SetShowText default True;
    property Font;
    property BorderStyle: TBorderStyle read FBorderStyle write SetBorderStyle default bsSingle;
    property ForeColor: TColor read FForeColor write SetForeColor default clBlack;
    property BackColor: TColor read FBackColor write SetBackColor default clWhite;
    property Kind : TBarKind read FKind write SetKind default bkHorizontal;
    property MinValue: Longint read FMinValue write SetMinValue default 0;
    property MaxValue: Longint read FMaxValue write SetMaxValue default 100;
    property ParentFont;
    property ParentShowHint;
    property Progress: Longint read FCurValue write SetProgress;
    property ShowHint;
    property Visible;
  end;

type
  TSpeedStyle = (sbSpeed, sbBitBtn, sbWin95);

  TCalStyleSpeed = class(TSpeedButton)
  private
    FStyle : TSpeedStyle;
    procedure SetStyle(Value : TSpeedStyle);
  protected
    procedure Paint; override;
  published
    property Style : TSpeedStyle read FStyle write SetStyle default sbSpeed;
  end;


type
  PBoolean = ^Boolean;

  TCalCheckList = class(TCustomListbox)
  private
    { Private declarations }
    FRangeCheck : Boolean;
    FHints : TStrings;
    procedure Validate(n : Integer);
    procedure SetHints(value : TStrings);
  protected
    { Protected declarations }
    procedure DrawItem(Index: Integer; Rect: TRect; State: TOwnerDrawState); override;
  public
    { Public declarations }
    constructor Create(AOwner: TComponent); override;
    destructor Destroy; override;
    procedure SetData(const b: array of Boolean);
    procedure GetData(const p: array of PBoolean);
  published
    { Published declarations }
    property RangeCheck : Boolean read FRangeCheck write FRangeCheck default True;
    property Hints : TStrings read FHints write SetHints;
    property Align;
    property BorderStyle;
    property Color;
    property Columns;
    property Ctl3D;
    property DragCursor;
    property DragMode;
    property Enabled;
    property Font;
    property IntegralHeight;
    property ItemHeight;
    property Items;
    property ParentColor;
    property ParentCtl3D;
    property ParentFont;
    property ParentShowHint;
    property PopupMenu;
    property ShowHint;
    property Sorted;
    property TabOrder;
    property TabStop;
    property Visible;
    property OnClick;
    property OnDblClick;
    property OnDragDrop;
    property OnDragOver;
    property OnEndDrag;
    property OnEnter;
    property OnExit;
    property OnKeyDown;
    property OnKeyPress;
    property OnKeyUp;
    property OnMouseDown;
    property OnMouseMove;
    property OnMouseUp;
  end;

  ECheckListError = class(Exception);

type
  TBooleanList = array[0..65528] of Boolean;
  PBooleanList = ^TBooleanList;

  EGridError = class(Exception);

  TGridSelectEvent = procedure (Sender : TObject; Index : Integer) of object;

  TCellSelectedEvent = procedure (Sender : TObject; Index : Integer;
     IsSelected : Boolean) of object;

  TMultiDrawCellEvent = procedure (Sender : TObject; Index: Integer; Rect : TRect;
     State : TGridDrawState) of object;

  TMultiSelectCellEvent = procedure (Sender : TObject; Index: Integer;
     var CanSelect: Boolean) of object;


  TCalMultiGrid = class(TCustomGrid)
  private
    { Private declarations }
    FSelected       : PBooleanList;
    FSelCount       : Integer;
    FSelColor       : TColor;
    FMulti          : Boolean;
    FAllowMulti     : Boolean;
    FOnSelect       : TGridSelectEvent;
    FOnCellSelected : TCellSelectedEvent;
    FOnDrawCell     : TMultiDrawCellEvent;
    FOnSelectCell   : TMultiSelectCellEvent;
    FOnTopLeftChange: TNotifyEvent;
    FUpdates        : Integer;
    FLimit          : Integer;
    FDropFocus      : Integer;
    function GetSelected(i : Integer): Boolean;
    procedure SetSelected(i : Integer; Sel : Boolean);
    function GetFocus : Integer;
    procedure SetFocus(i : Integer);
    procedure SetMulti(m: Boolean);
    procedure SetSelColor(value: TColor);
    function GetThumbTrack: Boolean;
    procedure SetThumbTrack(value : Boolean);
    procedure SetDropFocus(value: Integer);
  protected
    { Protected declarations }
    procedure SizeChanged(OldColCount, OldRowCount: Longint); override;
    procedure DrawCell(ACol, ARow: Longint; ARect: TRect;
       AState: TGridDrawState); override;
    procedure CellSelected(i : Integer; IsSelected : Boolean); virtual;
    function SelectCell(ACol, ARow: Longint): Boolean; override;
    procedure TopLeftChanged; override;
    procedure BeginUpdate;
    procedure EndUpdate;
  public
    { Public declarations }
    constructor Create(AOwner : TComponent); override;
    destructor Destroy; override;
    procedure SelectAll;
    procedure DeselectAll;
    function CellIndex(ACol, ARow : Longint) : Integer;
    procedure Reset;
    procedure SetSize(AColCount, ARowCount : Longint);
    procedure SizeGrid;
    procedure Select(Index: Integer);
    function MouseToCell(X, Y: Integer): Integer;
    function CellBounds(i: Integer): TRect;

    property SelCount : Integer read FSelCount;
    property Selected[i: Integer] : Boolean read GetSelected write SetSelected;
    property Multi : Boolean read FMulti write SetMulti;
    property DropFocus: Integer read FDropFocus write SetDropFocus;
    property Canvas;
    property TopRow;
    property LeftCol;
    property VisibleRowCount;
    property VisibleColCount;
  published
    { Published declarations }
    property Focus : Integer read GetFocus write SetFocus;
    property OnSelect : TGridSelectEvent read FOnSelect write FOnSelect;
    property OnCellSelected : TCellSelectedEvent read FOnCellSelected write FOnCellSelected;
    property OnDrawCell : TMultiDrawCellEvent read FOnDrawCell write FOnDrawCell;
    property OnSelectCell : TMultiSelectCellEvent read FOnSelectCell write FOnSelectCell;
    property OnTopLeftChange : TNotifyEvent read FOnTopLeftChange write FOnTopLeftChange;
    property AllowMulti: Boolean read FAllowMulti write FAllowMulti;
    property Limit : Integer read FLimit write FLimit;
    property SelColor : TColor read FSelColor write SetSelColor;
    property ThumbTrack : Boolean read GetThumbTrack write SetThumbTrack default False;
    property DefaultColWidth;
    property DefaultRowHeight;
    property RowCount;
    property ColCount;
    property Color;
    property Ctl3D;
    property DefaultDrawing;
    property DragCursor;
    property DragMode;
    property Enabled;
    property Font;
    property GridLineWidth;
    property ParentColor;
    property ParentCtl3D;
    property ParentFont;
    property ParentShowHint;
    property PopupMenu;
    property ShowHint;
    property Scrollbars;
    property TabOrder;
    property Visible;
    property OnClick;
    property OnDblClick;
    property OnDragDrop;
    property OnDragOver;
    property OnEndDrag;
    property OnEnter;
    property OnExit;
    property OnKeyDown;
    property OnKeyPress;
    property OnKeyUp;
    property OnMouseDown;
    property OnMouseMove;
    property OnMouseUp;
  end;


type
  TDragDrop = class(TComponent)
  private
    { Private declarations }
    FFiles : TStrings;
    FAutoClear : Boolean;
    procedure SetFiles(list: TStrings);
  protected
    { Protected declarations }
  public
    { Public declarations }
    constructor Create(AOwner: TComponent); override;
    destructor Destroy; override;
    property Files : TStrings read FFiles write SetFiles;

  published
    { Published declarations }
    property AutoClear : Boolean read FAutoClear write FAutoClear default True;
  end;

type
  TCalDropClient = class(TDragDrop)
  private
    { Private declarations }
    FDropPos     : TPoint;
    FOnDropFiles : TNotifyEvent;
    FHandle      : HWND;
  protected
    { Protected declarations }
    procedure Loaded; override;
    procedure ExtractFiles(Drop: THandle);
  public
    { Public declarations }
    constructor Create(AOwner: TComponent); override;
    destructor Destroy; override;
    class procedure CheckMessage(var Msg: TMsg; var Handled: Boolean);
    property  DropPos: TPoint read FDropPos;
  published
    { Published declarations }
    property OnDropFiles: TNotifyEvent read FOnDropFiles write FOnDropFiles;
    property Handle : HWND read FHandle;
  end;

type
  TFileDragEvent = procedure (Sender : TObject; X, Y: Integer;
    Target : HWND; var Accept : Boolean) of object;

  TFileDropEvent = procedure (Sender : TObject; X, Y: Integer;
    Target : HWND) of object;

  TCalDropServerError = class(Exception);

  TCalDropServer = class(TDragDrop)
  private
    { Private declarations }
    DropPoint       : TPoint;
    AllowDrop       : Boolean;
    DropWindow      : HWND;
    FInternalDrop   : Boolean;
    FDesktopDrop    : Boolean;
    FOnFileDrag     : TFileDragEvent;
    FOnFileDrop     : TFileDropEvent;
    FOnDeskDrop     : TFileDropEvent;
  protected
    { Protected declarations }
  public
    { Public declarations }
    procedure DropFiles(Wnd: HWnd; AMousePos: TPoint);
    function CanDrop: Boolean;
    procedure DragFinished;
  published
    { Published declarations }
    property InternalDrop : Boolean read FInternalDrop write FInternalDrop default False;
    property DesktopDrop : Boolean read FDesktopDrop write FDesktopDrop default False;
    property OnFileDrag : TFileDragEvent read FOnFileDrag write FOnFileDrag;
    property OnFileDrop : TFileDropEvent read FOnFileDrop write FOnFileDrop;
    property OnDeskDrop : TFileDropEvent read FOnDeskDrop write FOnDeskDrop;
  end;

type
  TCalSystemMenu = class(TComponent)
  private
    { Private declarations }
    FHandle : HMenu;
  protected
    { Protected declarations }
    procedure Loaded; override;
  public
    { Public declarations }
    procedure Reset;
    procedure Add(Caption: PChar; Command: Word);
    procedure Insert(Index: Integer; Caption: PChar; Command: Word);
    procedure AddSeparator;
    procedure Delete(Index: Integer);
    procedure DeleteCommand(Command: Word);
    procedure Rename(Command: Word; Caption: PChar);
    property Handle: HMenu read FHandle;
  published
    { Published declarations }
  end;

type
  TCalFormDragState = (fdNone, fdSolidMove, fdSolidSize, fdHollowMove, fdHollowSize);

  TCalFormDrag = class(TComponent)
  private
    { Private declarations }
    FMouseMove : TMouseMoveEvent;
    FMouseDown : TMouseEvent;
    FMouseUp   : TMouseEvent;
    FPaint     : TNotifyEvent;
    FResize    : TNotifyEvent;
    FDragState : TCalFormDragState;
    FDragWidth : Integer;
    FMinWidth  : Integer;
    FMinHeight : Integer;
    FMaxWidth  : Integer;
    FMaxHeight : Integer;
    FAllowMove : Boolean;
    FAllowSize : Boolean;
    FHollow    : Boolean;
    xofs, yofs : Integer;
    FormRect   : TRect;
    procedure HandleMouseMove(Sender: TObject; Shift: TShiftState; X, Y: Integer);
    procedure HandleMouseDown(Sender: TObject; Button: TMouseButton;
      Shift: TShiftState; X, Y: Integer);
    procedure HandleMouseUp(Sender: TObject; Button: TMouseButton;
      Shift: TShiftState; X, Y: Integer);
  protected
    { Protected declarations }
    procedure Loaded; override;
    procedure StartHollowDrag;
    procedure InverseRect;
  public
    { Public declarations }
    constructor Create(AOwner: TComponent); override;
    property DragState : TCalFormDragState read FDragState;
  published
    { Published declarations }
    property DragWidth : Integer read FDragWidth write FDragWidth default 16;
    property Hollow    : Boolean read FHollow write FHollow default False;
    property MinWidth  : Integer read FMinWidth write FMinWidth default 0;
    property MinHeight : Integer read FMinHeight write FMinHeight default 0;
    property MaxWidth  : Integer read FMaxWidth write FMaxWidth default High(Integer);
    property MaxHeight : Integer read FMaxHeight write FMaxHeight default High(Integer);
    property AllowMove : Boolean read FAllowMove write FAllowMove default True;
    property AllowSize : Boolean read FAllowSize write FAllowSize default True;
  end;

const
  MsgDialogSounds : Boolean = False;

var
  ApplicationPath : TFilename;
  WinPath : TFilename;

function Min(a, b: Integer): Integer;
function Max(a, b: Integer): Integer;
{ Returns the smaller and larger of two values respectively }

function Range(n, lower, upper: Integer): Integer;
{ Constrains n to a lower and upper limit }

function Sign(x: Integer) : Integer;
{ Returns 1 if x > 0, -1 if x < 0 and 0 if x = 0 }

procedure Border3d(Canvas : TCanvas; Width, Height: Integer);
{ Draws a raised 3D border on a canvas, typically used in an
  OnPaint method of a TForm }

procedure ErrorMsg(const msg: string);
{ Displays a message dialog box indicating an error }

procedure PlaySound(const filename: TFilename);
{ Plays the specified WAV file as a sound effect.  If the filename
  is <None>, nothing is played }

function Intersects(const R, S: TRect): Boolean;
{ Returns True if the two rectangles intersect }

function NormalizeRect(p, q: TPoint): TRect;
{ Returns a rectangle defined by any two points.  When dragging a
  selection box with a mouse, the fixed corner and the moving
  corner may not always be top left and bottom right respectively.
  This function creates a valid TRect out of them }

function TimeStampToDate(FileDate: Longint): TDateTime;
{ Converts a DOS timestamp to TDateTime.  If the timestamp is invalid
  (some programs use invalid stamps as markers), the current date
  is returned instead of raising EConvertError }

function GetRegValue(key : string): string;
{ Returns a value from the Windows registration database, with the
  specified key from HKEY_CLASSES_ROOT }

function GetRadioIndex(const R: array of TRadioButton): Integer;
procedure SetRadioIndex(const R: array of TRadioBUtton; index: Integer);
function GetMenuCheck(const M: array of TMenuItem): Integer;
procedure SetMenuCheck(const M: array of TMenuItem; index: Integer);
{ These routines are useful for setting and querying the state of
  several controls.  Use them to simulate arrays and as an alternative
  to TRadioGroup. }

procedure RefreshCursor;
{ Updates the cursor image when you have changed the Cursor or DragCursor
  property of a control }

function AddHistory(Combo : TComboBox): Boolean;
{ Adds a combo box's Text string to its listbox, but only if the
  string is not empty and not already present in the list.  The item is
  inserted at the top of the list, and if there are more than 24 items,
  the bottom one is removed.  Returns true if the list is modified }

function MsgDialog(const Msg: string; AType: TMsgDlgType;
  AButtons: TMsgDlgButtons; HelpCtx: Longint): Word;
{ Calls the MessageDialog function, but also plays a suitable sound
  effect from the Control Panel settings.  The MsgDialogSounds variable
  enables the sounds }

function ShowModalDialog(FormClass : TFormClass): TModalResult;
{ A very simple way of displaying a dynamic modal form -- just pass the
  form's class name e.g. TForm1, and an instance will be created,
  shown as a modal dialog and then destroyed. }

function InitBitmap(ABitmap: TBitmap;
  AWidth, AHeight : Integer; Color : TColor) : TBitmap;
{ Initialises the bitmap's dimensions and fills it with the chosen colour }

procedure ShrinkIcon(H : HIcon; Glyph : TBitmap);
{ Shrinks a 32 x 32 icon down to a 16 x 16 bitmap }

const
  Uppers = ['A'..'Z'];
  Lowers = ['a'..'z'];
  Alphas = Uppers + Lowers;
  Digits = ['0'..'9'];
  AlphaDigits = Alphas + Digits;

  OneItem : array[Boolean] of string[2] = ('s', '');

type
  TAttrStr = string[5];

  TUniqueStrings = class(TStringList)
    constructor Create;
  end;

function LowCase(c : Char) : Char;
{ converts a character to lower case }

function InString(c: char; const s: string) : Boolean;
{ Returns true if a character is present in a string.
  Probably faster than Pos. }

function CharCount(c: Char; const S: string): Integer;
{ Returns the number of occurences of c in S }

function Blank(const s: string): boolean;
{ Returns true if the string is empty or consists of spaces }

function MakePath(const s: string): string;
{ Adds a trailing backslash to a directory name, if necessary }

function MakeDirname(const s: string): string;
{ Removes a trailing backslash from a directory name, if necessary }

function ExtractFileDir(const s: string): string;
{ Calls MakeDirname after calling ExtractFilePath }

function FormatByte(size : Longint): string;
{ Formats a number (assumed to be bytes) to display as bytes,
  KB or MB, for example "245 bytes", "1.60KB", "44.10MB" }

function GetStrKey(const s: string): string;
function GetStrValue(const s: string): string;
{ Returns the left and right sides, respectively, of a string with the
  structure Key=Value }

function SetStrValue(const s, value: string): string;
{ If s contains an '=', the portion to the right of '=' is set
  to the value }

function FillString(c: char; n: Byte): string;
{ Returns a string of length n containing only the specified character }

function Unformat(const source, pattern: string; const args: array of const): Integer;
{ The opposite of Format, Unformat splits up a formatted source string
  into substrings and Integers.  It is an alternative to parsing when
  the format is known to be fixed.  The pattern parameter contains the format
  string, which is a combination of plain characters and format specifiers.

  The following specifiers are supported:

  %s   indicates that a string value is required
  %d   indicates that an integer value is required
  %S   indicates that a string value should be ignored
  %D   indicates that an integer value should be ignored

  Unformat compares the source with the pattern, and plain characters
  that do not match will raise an EConvertError.  When a format specifier
  is encountered in the pattern, an argument is fetched and used to
  store the result that is obtained from the source.  Then the comparison
  continues.

  For each %s, the args list must contain a pointer to a string variable,
  followed by an integer specifying the maximum length of the string.
  For each %d, the args list must contain a pointer to an integer variable.

  When the end of the source string is reached, the function returns
  without modifying the remaining arguments, so you might wish to initialize
  your variables to "default" values before the function call.

  Unformat returns the number of values it has extracted.

  Examples:

  var
    s1, s2: string[31];
    i : Integer;

  Unformat('[abc]123(def)', '[%s]%d(%s)', [@s1, 31, @i, @s2, 31]);
    (* s1 = 'abc', i = 123, s2 = 'def' *)

  Unformat('Hello, Universe!!!', '%s, %s%d', [@s1, 31, @s2, 31, @i]);
    (* s1 = 'Hello', s2 = 'Universe!!!', i is untouched *)

  Unformat('How much wood could a woodchuck chuck...',
           '%S %S %s could a %S %s...', [@s1, 31, @s2, 31]);
    (* s1 = 'wood', s2 = 'chuck' *)
}


function FileParams(files: TStrings): string;
{ Assumes that the strings parameter contains a list of filename, and
  concatenates the names to form a single string suitable for passing
  as a command line parameter.  Filenames with no extension have an
  extra '.' appended to ensure correct interpretation }

function GetWord(var s: OpenString): string;
{ Skips spaces and returns the next word in a string.  The word
  is deleted from the string }

function StringAsPChar(var s: OpenString): PChar;
{ Modifies a string so that it can be used as a PChar without additional
}

function AttrToStr(attr : Integer): TAttrStr;
function LTrim(const s: string): string;
function RTrim(const s: string): string;
function Trim(const s: string): string;


var
  IconSelForm: TIconSelForm;

procedure Register;

implementation

{$R *.DFM}

uses ShellAPI, ExtCtrls, MMSystem, IniFiles;

procedure TIconSelForm.SearchFile(const filename: TFilename);
var
  i : Integer;
  h : THandle;
  s : array[0..79] of Char;
begin
  FileEdit.Text := Lowercase(filename);
  Update;

  h := ExtractIcon(HInstance, StrPCopy(s, filename), 0);

  if h <= 1 then begin
    StrPCopy(s, Application.ExeName);
    MessageDlg(Format('There are no icons in this file.  Please choose one ' +
      'from %s', [ExtractFilename(Application.ExeName)]),
      mtInformation, [mbOK], 0);
    FileEdit.Text := Lowercase(Application.ExeName);
  end
  else DestroyIcon(h);

  Screen.Cursor := crHourGlass;
  try
    for i := 0 to Icons.Count-1 do DestroyIcon(Word(Icons[i]));
    Icons.Clear;

    i := 0;
    h := ExtractIcon(HInstance, s, i);
    while h > 1 do begin
      Icons.Add(Pointer(h));
      Inc(i);
      h := ExtractIcon(HInstance, s, i);
    end;

    Grid.ColCount := Icons.Count;
    if Index >= Icons.Count then Index := Icons.Count-1;
    Grid.LeftCol := Index;
    Grid.Col := Grid.LeftCol;
    Index := 0;
  finally
    Screen.Cursor := crDefault;
  end;
  Grid.Invalidate;
end;


procedure TIconSelForm.FormCreate(Sender: TObject);
begin
  Icons := TList.Create;
  Index := 0;
end;

procedure TIconSelForm.FormDestroy(Sender: TObject);
var i: Integer;
begin
  for i := 0 to Icons.Count-1 do DestroyIcon(Word(Icons[i]));
end;

procedure TIconSelForm.GridDrawCell(Sender: TObject; Col, Row: Longint;
  Rect: TRect; State: TGridDrawState);
begin
  if Col < Icons.Count then
    DrawIcon(Grid.Canvas.Handle, Rect.Left + 4, Rect.Top + 4, HIcon(Icons[Col]));
end;

procedure TIconSelForm.GridSelectCell(Sender: TObject; Col, Row: Longint;
  var CanSelect: Boolean);
begin
  CanSelect := Col < Icons.Count;
end;

procedure TIconSelForm.FormShow(Sender: TObject);
begin
  if FileEdit.Text = '' then begin
    FileEdit.Text := Application.ExeName;
    Index := 0;
  end;
  SearchFile(FileEdit.Text);
end;

procedure TIconSelForm.BrowseBtnClick(Sender: TObject);
begin
  with OpenDialog do begin
    Filename := FileEdit.Text;
    if Execute then SearchFile(Filename);
  end;
end;

constructor TCalIconDialog.Create(AOwner : TComponent);
begin
  inherited Create(AOwner);
  FFilename := Application.Exename;
  FIconIndex := 0;
  FHistory := TStringList.Create;
end;

destructor TCalIconDialog.Destroy;
begin
  FHistory.Free;
  inherited Destroy;
end;

procedure TCalIconDialog.SetHistory(value : TStrings);
begin
  FHistory.Assign(value);
end;


function TCalIconDialog.Execute : Boolean;
var form : TIconSelForm;
begin
  form := TIconSelForm.Create(Application);
  with Form do
  try
    FileEdit.Text := Filename;
    OpenDialog.HistoryList.Assign(FHistory);
    Index := IconIndex;

    if ShowModal = mrOK then begin
      Result := True;
      Filename := Lowercase(FileEdit.Text);
      IconIndex := Grid.Col;
      if FHistory.IndexOf(Filename) = -1 then FHistory.Insert(0, Filename);
      if FHistory.Count > 24 then FHistory.Delete(24);
    end
    else Result := False;
  finally
    Free;
  end;
end;

constructor TCalAppHolder.Create(AOwner : TComponent);
begin
  inherited Create(AOwner);
  FHintColor := $0080FFFF;
  FHintPause := 800;
end;


destructor TCalAppHolder.Destroy;
begin
  if Assigned(FOnWndProc) then Application.UnHookMainWindow(FOnWndProc);
  inherited Destroy;
end;


procedure TCalAppHolder.Loaded;
begin
  with Application do begin
    if FHelpFile > '' then HelpFile := ExtractFilePath(ExeName) + FHelpFile;
    HintColor := FHintColor;
    HintPause := FHintPause;
    if FTitle > '' then Title := FTitle;
    OnActivate := FOnActivate;
    OnDeactivate := FOnDeactivate;
    OnException := FOnException;
    OnHelp := FOnHelp;
    OnHint := FOnHint;
    OnIdle := FOnIdle;
    OnMessage := FOnMessage;
    OnMinimize := FOnMinimize;
    OnRestore := FOnRestore;
    OnShowHint := FOnShowHint;
    if Assigned(FOnWndProc) then HookMainWindow(FOnWndProc);
  end;

  with Screen do begin
    OnActiveControlChange := FOnActiveControlChange;
    OnActiveFormChange := FOnActiveFormChange;
  end;
end;


constructor TCalBarGauge.Create(AOwner: TComponent);
begin
  inherited Create(AOwner);
  ControlStyle := ControlStyle + [csFramed, csOpaque];
  { default values }
  FMinValue := 0;
  FMaxValue := 100;
  FCurValue := 0;
  FShowText := True;
  FBorderStyle := bsSingle;
  FForeColor := clBlack;
  FBackColor := clWhite;
  FCtl3D := True;
  Width := 100;
  Height := 100;
end;


function TCalBarGauge.GetPercentDone: Integer;
begin
  if (FMaxValue = FMinValue) or (FCurValue = FMinValue) then Result := 0
  else Result := Trunc((FCurValue - FMinValue) / (FMaxValue - FMinValue) * 100);
end;


procedure TCalBarGauge.SetShowText(Value: Boolean);
begin
  if Value <> FShowText then begin
    FShowText := Value;
    Repaint;
  end;
end;

procedure TCalBarGauge.SetBorderStyle(Value: TBorderStyle);
begin
  if Value <> FBorderStyle then begin
    FBorderStyle := Value;
    Repaint;
  end;
end;

procedure TCalBarGauge.SetForeColor(Value: TColor);
begin
  if Value <> FForeColor then begin
    FForeColor := Value;
    Repaint;
  end;
end;

procedure TCalBarGauge.SetBackColor(Value: TColor);
begin
  if Value <> FBackColor then begin
    FBackColor := Value;
    Repaint;
  end;
end;

procedure TCalBarGauge.SetMinValue(Value: Longint);
begin
  if Value <> FMinValue then begin
    FMinValue := Value;
    Repaint;
  end;
end;

procedure TCalBarGauge.SetMaxValue(Value: Longint);
begin
  if Value <> FMaxValue then begin
    FMaxValue := Value;
    Repaint;
  end;
end;


procedure TCalBarGauge.SetProgress(Value: Longint);
var PrevPercent : Integer;
begin
  if (FCurValue <> Value) and (Value >= FMinValue) and (Value <= FMaxValue) then begin
    PrevPercent := GetPercentDone;
    FCurValue := Value;
    if GetPercentDone <> PrevPercent then Repaint;
  end;
end;


procedure TCalBarGauge.SetCtl3D(Value: Boolean);
begin
  if Value <> FCtl3D then begin
    FCtl3D := Value;
    Repaint;
  end;
end;


procedure TCalBarGauge.SetKind(Value: TBarKind);
begin
  if Value <> FKind then begin
    FKind := Value;
    Repaint;
  end;
end;


procedure TCalBarGauge.AddProgress(Value: Longint);
begin
  Progress := FCurValue + Value;
end;



procedure TCalBarGauge.Paint;
var
  r: TRect;
  x, y: Integer;
  s: string[4];
begin
  r := Rect(0, 0, Width, Height);
  with Canvas do begin

    if BorderStyle = bsSingle then begin
      if Ctl3D then Frame3D(Canvas, r, clBtnShadow, clBtnHighlight, 1);
      Frame3D(Canvas, r, clBlack, clBlack, 1);
    end;

    if Kind = bkHorizontal then begin
      x := MulDiv(r.Right - r.Left, PercentDone, 100);
      Brush.Color := ForeColor;
      FillRect(Rect(r.Left, r.Top, r.Left + x, r.Bottom));
      Brush.Color := BackColor;
      FillRect(Rect(r.Left + x, r.Top, r.Right, r.Bottom));
    end
    else begin
      y := MulDiv(r.Bottom - r.Top, PercentDone, 100);
      Brush.Color := ForeColor;
      FillRect(Rect(r.Left, r.Bottom - y, r.Right, r.Bottom));
      Brush.Color := BackColor;
      FillRect(Rect(r.Left, r.Top, r.Right, r.Bottom - y));
    end;

    if ShowText then begin
      s := Format('%d%%', [PercentDone]);
      Brush.Style := bsClear;
      Font.Assign(Self.Font);
      with r do begin
        x := (Width + 1 - TextWidth(s)) div 2;
        y := (Height + 1 - TextHeight(s)) div 2;
      end;
      TextRect(r, x, y, S);
    end;
  end;
end;

procedure TCalStyleSpeed.SetStyle(Value : TSpeedStyle);
begin
  if Value <> FStyle then begin
    FStyle := Value;
    Invalidate;
  end;
end;


procedure TCalStyleSpeed.Paint;
begin
  inherited Paint;

  if Style = sbSpeed then Exit;

  if Style = sbBitBtn then
    with Canvas do begin
      Pen.Color := clBtnShadow;
      MoveTo(0, Height-2);
      LineTo(0, 0);
      LineTo(Width-2, 0);

      if FState in [bsDown, bsExclusive] then begin
        LineTo(Width-2, Height-2);
        LineTo(0, Height-2);

        Pen.Color := clBlack;
        MoveTo(1, Height-2);
        LineTo(1, 1);
        LineTo(Width-2, 1);

        Pen.Color := clWhite;
        MoveTo(Width-1, 0);
        LineTo(Width-1, Height-1);
        LineTo(-1, Height-1);
      end
    end

  else
    with Canvas do
    if FState in [bsDown, bsExclusive] then begin
      Pen.Color := clWhite;
      MoveTo(Width-1, 0);
      LineTo(Width-1, Height-1);
      LineTo(-1, Height-1);
    end
    else begin
      Pen.Color := clWhite;
      MoveTo(0, Height-1);
      LineTo(0, 0);
      LineTo(Width-1, 0);

      Pen.Color := clBtnFace;
      MoveTo(1, Height-2);
      LineTo(1, 1);
      LineTo(Width-2, 1);

      Pixels[0, Height-1] := clBlack;
      Pixels[1, Height-2] := clBtnShadow;
    end;
end;

{$R calmuniv.RES}

var
  CheckedBmp, UncheckedBmp : TBitmap;


procedure LoadCheckboxBitmaps;
begin
  CheckedBmp := TBitmap.Create;
  CheckedBmp.Handle := LoadBitmap(HInstance, 'CHECKLISTCHECKED');
  UncheckedBmp := TBitmap.Create;
  UncheckedBmp.Handle := LoadBitmap(HInstance, 'CHECKLISTUNCHECKED');
end;

constructor TCalCheckList.Create(AOwner: TComponent);
begin
  inherited Create(AOwner);
  Color := clBtnFace;
  IntegralHeight := True;
  Style := lbOwnerDrawFixed;
  MultiSelect := True;
  ExtendedSelect := False;
  ItemHeight := 20;
  FRangeCheck := True;
  FHints := TStringList.Create;
  if CheckedBmp = nil then LoadCheckboxBitmaps;
end;


destructor TCalCheckList.Destroy;
begin
  FHints.Free;
  inherited Destroy;
end;

procedure TCalCheckList.SetHints(value: TStrings);
begin
  FHints.Assign(value);
end;


procedure TCalCheckList.DrawItem(Index: Integer; Rect: TRect; State: TOwnerDrawState);
var
  bmp : TBitmap;
  h : Integer;
begin
  h := Rect.Bottom - Rect.Top;

  Canvas.Brush.Color := Color;
  Canvas.FillRect(Rect);

  if Selected[Index] then
    bmp := CheckedBmp else bmp := UncheckedBmp;

  Canvas.Draw(4, Rect.Top + (h - bmp.Height) div 2, bmp);
  Canvas.TextOut(Rect.Left + 22,
    Rect.Top + (h - Abs(Font.Height)) div 2 - 1, Items[Index]);
end;


procedure TCalCheckList.Validate(n : Integer);
begin
  if FRangeCheck then
    if n < Items.Count then
      raise ECheckListError.Create('Not enough elements in data array')
    else if n > Items.Count then
      raise ECheckListError.Create('Too many elements in data array')
end;


procedure TCalCheckList.SetData(const b: array of Boolean);
var
  i: Integer;
begin
  Validate(High(b)+1);
  i := 0;

  while (i <= High(b)) and (i < Items.Count) do begin
    Selected[i] := b[i];
    Inc(i);
  end;

  TopIndex := 0;
  Invalidate;
end;


procedure TCalCheckList.GetData(const p: array of PBoolean);
var
  i: Integer;
begin
  Validate(High(p)+1);
  i := 0;

  while (i <= High(p)) and (i < Items.Count) do begin
    p[i]^ := Selected[i];
    Inc(i);
  end;
end;


procedure DoneCheckList; far;
begin
  CheckedBmp.Free;
  UncheckedBmp.Free;
end;


constructor TCalMultiGrid.Create(AOwner : TComponent);
begin
  inherited Create(AOwner);
  FixedRows := 0;
  FixedCols := 0;
  DefaultDrawing := True;
  GridLineWidth := 0;
  Options := Options - [goRangeSelect];
  FDropFocus := -1;
  FMulti := False;
  FAllowMulti := True;
  FSelColor := clBtnFace;
  FSelected := AllocMem(RowCount * ColCount);
end;


destructor TCalMultiGrid.Destroy;
begin
  FreeMem(FSelected, RowCount * ColCount);
  inherited Destroy;
end;


function TCalMultiGrid.GetSelected(i : Integer): Boolean;
begin
  if (i >= 0) and (i < ColCount * RowCount) then Result := FSelected^[i]
  else raise EListError.Create('Index of out range');
end;


procedure TCalMultiGrid.SetSelected(i : Integer; Sel : Boolean);
begin
  if (i >= 0) and (i < ColCount * RowCount) then begin
    if FSelected^[i] <> Sel then begin
      FSelected^[i] := Sel;

      if Sel then begin
        Inc(FSelCount);
        if not FMulti and (FSelcount > 1) then begin
          FAllowMulti := True;
          FMulti := True;
        end;
      end
      else Dec(FSelCount);

      InvalidateCell(i mod ColCount, i div ColCount);
      if Assigned(FOnCellSelected) then FOnCellSelected(self, i, Sel);
    end
  end
  else raise EGridError.Create('Index of out range');
end;


{ BeginUpdate and EndUpdate

  These are internal methods used to prevent the grid from redrawing
  when some shuffling of properties is taking place.  When TCalMultiGrid
  is in an "updating" state, OnSelectCell and OnDrawCell are bypassed }

procedure TCalMultiGrid.BeginUpdate;
begin
  Inc(FUpdates);
end;


procedure TCalMultiGrid.EndUpdate;
begin
  if FUpdates > 0 then Dec(FUpdates);
end;


function TCalMultiGrid.GetFocus : Integer;
begin
  Result := Row * ColCount + Col;
end;


procedure TCalMultiGrid.SetFocus(i : Integer);
begin
  if i < RowCount * ColCount then begin
    BeginUpdate;
    Row := i div ColCount;
    Col := i mod ColCount;
    EndUpdate;
  end;
end;


procedure TCalMultiGrid.SetMulti(m: Boolean);
begin
  if FMulti <> m then begin
    if m then begin
      FAllowMulti := True;
      FMulti := True;
    end
    else begin
      if SelCount > 0 then DeselectAll;
      FMulti := False;
    end;
  end;
end;


function TCalMultiGrid.CellBounds(i: Integer): TRect;
begin
  Result := CellRect(i mod ColCount, i div ColCount);
end;


procedure TCalMultiGrid.SetSelColor(value: TColor);
begin
  if FSelColor <> value then begin
    FSelColor := value;
    if SelCount > 0 then Invalidate;
  end;
end;

procedure TCalMultiGrid.SetSize(AColCount, ARowCount : Longint);
var
  f : Integer;
  p : PBooleanList;
  bufsize : Word;
begin
  if (AColCount = ColCount) and (ARowCount = RowCount) then exit;
  if AColCount = 0 then AColCount := 1;
  if ARowCount = 0 then ARowCount := 1;

  { The current selection is copied to a temporary buffer and then
    restored once the inherited sizing is complete }

  BeginUpdate;
  f := Focus;
  bufsize := Min(AColCount * ARowCount, ColCount * RowCount);
  p := AllocMem(bufsize);
  try
    Move(FSelected^, p^, bufsize);
    ColCount := AColCount;
    RowCount := ARowCount;
    Move(p^, FSelected^, bufsize);
    Focus := f;
  finally
    EndUpdate;
    FreeMem(p, bufsize);
    Invalidate;
  end;
end;


procedure TCalMultiGrid.SizeGrid;
var c, r: Longint;
begin
  { try to display without the scroll bar first }

  c := Width div DefaultColWidth;
  if c = 0 then Inc(c);
  r := Limit div c;
  if Limit mod c > 0 then Inc(r);

  { if the computed row count exceeds the number of rows that
    can be displayed, take the scroll bar width into account and recalculate }

  if (Height - 4) div DefaultRowHeight < r then begin
    c := (Width - GetSystemMetrics(SM_CXVSCROLL)) div DefaultColWidth;
    if c = 0 then Inc(c);
    r := Limit div c;
    if Limit mod c > 0 then Inc(r);
  end;

  Setsize(c, r);
end;



procedure TCalMultiGrid.SizeChanged(OldColCount, OldRowCount: Longint);
begin
  inherited SizeChanged(OldColCount, OldRowCount);
  FreeMem(FSelected, OldColCount * OldRowCount);
  FSelected := AllocMem(ColCount * RowCount);
end;


procedure TCalMultiGrid.CellSelected(i : Integer; IsSelected : Boolean);
begin
  Selected[i] := IsSelected;
  InvalidateCell(i mod ColCount, i div ColCount);
  if Assigned(FOnCellSelected) then FOnCellSelected(self, i, IsSelected);
end;


function TCalMultiGrid.CellIndex(ACol, ARow : Longint) : Integer;
begin
  Result := ARow * ColCount + ACol;
end;


procedure TCalMultiGrid.Reset;
begin
  FillChar(FSelected^, ColCount * RowCount, False);
  FSelcount := 0;
  FMulti := False;
  Invalidate;
end;



function TCalMultiGrid.SelectCell(ACol, ARow: Longint): Boolean;
var
  i, j, index, lower, upper, temp: Integer;
begin
  if FUpdates > 0 then begin
    Result := True;
    exit;
  end;

  index := ARow * ColCount + ACol;

  Result := index < Limit;
  if Result and Assigned(FOnSelectCell) then
    FOnSelectCell(self, index, Result);

  if Result then begin

    if AllowMulti and (GetKeyState(VK_CONTROL) < 0) then begin
      { Ctrl-click.  Invert selection of target cell }
      FMulti := True;
      Selected[index] := not Selected[index];
    end
    else if AllowMulti and (GetKeyState(VK_SHIFT) < 0) then begin
      { Shift-click.  Select range of cells }
      FMulti := True;
      lower := Row * ColCount + Col;
      upper := index;
      if lower > upper then begin
        temp := lower;
        lower := upper;
        upper := temp;
      end;

      for i := 0 to Limit-1 do
        Selected[i] := (i >= lower) and (i <= upper);
    end
    else
      { normal click -- no Ctrl or Shift }
      if FMulti then begin
        if not FSelected^[index] then begin
          { turn off multi mode }
          FMulti := False;
          for i := 0 to Limit-1 do Selected[i] := False;
          Selected[index] := True;
        end;
      end
      else begin
        { change highlighted cell }
        i := Row * ColCount + Col;
        Selected[i] := False;
        Selected[index] := True;
      end;
    if Assigned(FOnSelect) then FOnSelect(self, index);
  end;
end;


procedure TCalMultiGrid.Select(Index : Integer);
var ACol, ARow, c, r: Longint;
begin
  c := Col; r := Row;
  if SelectCell(Index mod ColCount, Index div ColCount) then begin
    Focus := Index;
    InvalidateCell(c, r);
    Update;
  end;
end;


procedure TCalMultiGrid.DrawCell(ACol, ARow: Longint; ARect: TRect;
   AState: TGridDrawState);
var i : Integer;
begin
  if FUpdates > 0 then exit;
  i := ARow * ColCount + ACol;
  with Canvas do begin
    if FSelected^[i] then begin
      Brush.Color := SelColor;
      Include(AState, gdSelected);
    end
    else begin
      Brush.Color := Color;
      Exclude(AState, gdSelected);
    end;
    if DefaultDrawing then begin
      if gdFocused in AState then DrawFocusRect(ARect);
      FillRect(ARect);
    end;
  end;
  Exclude(AState, gdFixed);
  if (i < Limit) and Assigned(FOnDrawCell) then
    FOnDrawCell(self, i, ARect, AState);
end;


procedure TCalMultiGrid.DeselectAll;
var i: Integer;
begin
  for i := 0 to Limit-1 do Selected[i] := False;
  if Assigned(FOnSelect) then FOnSelect(self, Focus);
end;


procedure TCalMultiGrid.SelectAll;
var i: Integer;
begin
  for i := 0 to Limit-1 do Selected[i] := True;
  if Assigned(FOnSelect) then FOnSelect(self, Focus);
end;


function TCalMultiGrid.MouseToCell(X, Y: Integer): Integer;
begin
  with MouseCoord(X, Y) do Result := Y * ColCount + X;
end;

procedure TCalMultiGrid.TopLeftChanged;
begin
  if Assigned(FOnTopLeftChange) then FOnTopLeftChange(self);
end;

function TCalMultiGrid.GetThumbTrack: Boolean;
begin
  Result := goThumbTracking in Options;
end;

procedure TCalMultiGrid.SetThumbTrack(value : Boolean);
begin
  if value then Options := Options + [goThumbTracking]
  else Options := Options - [goThumbTracking];
end;


procedure TCalMultiGrid.SetDropFocus(value: Integer);
begin
  if FDropFocus <> Value then begin
    if FDropFocus <> -1 then
      Canvas.DrawFocusRect(CellBounds(FDropFocus));
    if value <> -1 then
      Canvas.DrawFocusRect(CellBounds(value));

    FDropFocus := value;
  end;
end;

var Clients : TList;

constructor TCalDropClient.Create(AOwner: TComponent);
begin
  if not (AOwner is TWinControl) then
    raise EInvalidOperation.Create('A TCalDropClient must be owned by a windowed control');
  inherited Create(AOwner);
  Clients.Add(self);
end;


destructor TCalDropClient.Destroy;
begin
  Clients.Remove(self);
  inherited Destroy;
end;


procedure TCalDropClient.ExtractFiles(drop: THandle);
var
  fcount : Integer;
  i      : Integer;
  s      : array[0..127] of Char;
begin
  if AutoClear then Files.Clear;
  fcount := DragQueryFile(drop, Word(-1), @s, 127);
  for i := 0 to fcount-1 do begin
    DragQueryFile(drop, i, @s, 127);
    Files.Add(Lowercase(StrPas(@s)));
  end;
  DragQueryPoint(drop, FDropPos);
  DragFinish(drop);
  if Assigned(FOnDropFiles) then FOnDropFiles(self);
end;


procedure TCalDropClient.Loaded;
begin
  inherited Loaded;
  FHandle := (Owner as TWinControl).Handle;
  DragAcceptFiles(FHandle, True);
end;


class procedure TCalDropClient.CheckMessage(var Msg: TMsg; var Handled: Boolean);
var
  i: Integer;
  L: PPointerList;
begin
  if Msg.Message = WM_DROPFILES then begin
    L := Clients.List;
    for i := 0 to Clients.Count-1 do
      with TCalDropClient(L^[i]) do
      if Handle = Msg.HWnd then begin
        try
          ExtractFiles(Msg.wParam);
        except
          on E: Exception do Application.HandleException(E);
        end;
        Handled := True;
        Exit;
      end;
  end;
end;

constructor TDragDrop.Create(AOwner: TComponent);
begin
  inherited Create(AOwner);
  FFiles := TStringList.Create;
  FAutoClear := True;
end;

destructor TDragDrop.Destroy;
begin
  FFiles.Free;
  inherited Destroy;
end;

procedure TDragDrop.SetFiles(list: TStrings);
begin
  FFiles.Assign(list);
end;

type
  { Windows expects a WM_DROPFILES message to contain a memory
    handle with this structure as a header }

  PDropFileRec = ^TDropFileRec;
  TDropFileRec = record
    Size: Word;
    MousePos: TPoint;
    InNonClientArea: Boolean;
  end;

procedure TCalDropServer.DropFiles(Wnd: HWnd; AMousePos: TPoint);
var
  H: THandle;
  p: PDropFileRec;
  data : PChar;
  size : Word;
  i, count: Integer;
begin
  { Thanks to Brian Andersen for this procedure's algorithm (it doesn't
    seem to be documented by Microsoft) }

  if GetWindowLong(Wnd, GWL_EXSTYLE) and WS_EX_ACCEPTFILES = 0 then
    raise TCalDropServerError.Create('This windows does not accept dropped files');

  WinProcs.ScreenToClient(Wnd, AMousePos);

  { Find the size of the buffer needed to hold the filenames }
  size := SizeOf(TDropFileRec) + 1;
  count := 0;
  while (size < 65276) and (count < Files.Count) do begin
    Inc(size, Length(Files[count]) + 1);
    Inc(count);
  end;

  { Allocate the buffer }
  H := GlobalAlloc(GMEM_SHARE or GMEM_MOVEABLE or GMEM_ZEROINIT, size);

  if H = 0 then raise TCalDropServerError.Create('Unable to allocate file drop buffer.');

  p := GlobalLock(H);

  { Initialize the header info }
  with p^ do begin
    Size := SizeOf(TDropFileRec);
    MousePos := AMousePos;
    InNonClientArea := SendMessage(Wnd, WM_NCHITTEST, 0, Longint(AMousePos)) <> HTCLIENT;
  end;

  { Append the filenames to the buffer after advancing the pointer
    beyond the header }
  data := PChar(p) + Sizeof(TDropFileRec);
  for i := 0 to count-1 do begin
     StrPCopy(data, Files[i]);
     Inc(data, Length(Files[i]) + 1);
  end;
  GlobalUnlock(H);

  { Drop the files }
  PostMessage(Wnd, WM_DROPFILES, H, 0);

  if AutoClear then Files.Clear;
end;


function TCalDropServer.CanDrop: Boolean;
var
  p: TPoint;
begin
  GetCursorPos(p);
  AllowDrop := False;
  DropWindow := WindowFromPoint(p);

  if ((DropWindow = GetDesktopWindow) and DesktopDrop) then begin
    AllowDrop := True;
    DropPoint := p;
  end
  else begin
    if ((DropWindow <> 0) and
      (InternalDrop or (GetWindowTask(DropWindow) <> GetCurrentTask)) and
      (GetWindowLong(DropWindow, GWL_EXSTYLE) and WS_EX_ACCEPTFILES > 0)) then begin
      AllowDrop := True;
      DropPoint := p;
      if Assigned(FOnFileDrag) then FOnFileDrag(self, p.x, p.y, DropWindow, AllowDrop);
    end;
  end;
  Result := AllowDrop;
end;


procedure TCalDropServer.DragFinished;
begin
  if AllowDrop then begin
    if DropWindow = GetDesktopWindow then begin
      if Assigned(FOnDeskDrop) then
        FOnDeskDrop(self, DropPoint.x, DropPoint.y, DropWindow);
    end
    else if Assigned(FOnFileDrop) then begin
      FOnFileDrop(self, DropPoint.x, DropPoint.y, DropWindow);
      DropFiles(DropWindow, DropPoint);
    end;
  end;
  AllowDrop := False;
end;

procedure TCalSystemMenu.Loaded;
begin
  inherited Loaded;
  FHandle := GetSystemMenu((Owner as TForm).Handle, False);
end;

procedure TCalSystemMenu.Reset;
begin
  FHandle := GetSystemMenu((Owner as TForm).Handle, True);
end;

procedure TCalSystemMenu.Add(Caption: PChar; Command: Word);
begin
  AppendMenu(FHandle, MF_ENABLED, Command, Caption);
end;

procedure TCalSystemMenu.AddSeparator;
begin
  AppendMenu(FHandle, MF_SEPARATOR, 0, nil);
end;


procedure TCalSystemMenu.Delete(Index: Integer);
begin
  DeleteMenu(FHandle, Index, MF_BYPOSITION);
end;

procedure TCalSystemMenu.DeleteCommand(Command: Word);
begin
  DeleteMenu(FHandle, Command, MF_BYCOMMAND);
end;

procedure TCalSystemMenu.Rename(Command: Word; Caption: PChar);
begin
  ModifyMenu(FHandle, Command, MF_BYCOMMAND, Command, Caption);
end;

procedure TCalSystemMenu.Insert(Index: Integer; Caption: PChar; Command: Word);
begin
  InsertMenu(FHandle, Index, MF_BYPOSITION, Command, Caption);
end;


var
  { GDI objects used for drawing the resize boxes }
  InversePen : HPen;

constructor TCalFormDrag.Create(AOwner: TComponent);
begin
  if not (AOwner is TForm) then
    raise EInvalidOperation.Create('TCalFormDrag can only be placed on a form');

  inherited Create(AOwner);
  FDragState := fdNone;
  FDragWidth := 16;
  FHollow    := False;
  FMaxWidth  := High(Integer);
  FMaxHeight := High(Integer);
  FAllowMove := True;
  FAllowSize := True;
end;


procedure TCalFormDrag.Loaded;
begin
  { The form's handlers must not be touched during design time,
    otherwise incorrect events will be saved }

  inherited Loaded;
  if not (csDesigning in ComponentState) then
    with Owner as TForm do begin
      { Save the form's original handlers... }
      FMouseMove := OnMouseMove;
      FMouseDown := OnMouseDown;
      FMouseUp   := OnMouseUp;
      FPaint     := OnPaint;
      FResize    := OnResize;

      { ...and make the form call our handlers instead }
      OnMouseMove := HandleMouseMove;
      OnMouseDown := HandleMouseDown;
      OnMouseUp   := HandleMouseUp;
    end;
end;


procedure TCalFormDrag.StartHollowDrag;
begin
  { Create a suitable pen and set up the screen device context
    ready for drawing }
  InversePen := CreatePen(PS_SOLID, 1, ColorToRGB(clBlack));
  { Obtain and draw the initial rectangle }
  FormRect := TForm(Owner).BoundsRect;
  InverseRect;
end;


procedure TCalFormDrag.InverseRect;
var
  ScreenDC: HDC;
  OldPen     : HPen;
  OldRop2    : Integer;
begin
  { Draws an inverse rectangle on the screen device context, the
    coordinates specified by FormRect.  The rectangle is two
    pixels thick but it is faster to draw two thin ones than to
    use a thick pen }

  ScreenDC := GetDC(0);
  OldRop2 := SetRop2(ScreenDC, R2_NOT);
  OldPen := SelectObject(ScreenDC, InversePen);

  with FormRect do begin
    MoveTo(ScreenDC, Left, Top);
    LineTo(ScreenDC, Right-1, Top);
    LineTo(ScreenDC, Right-1, Bottom-1);
    LineTo(ScreenDC, Left, Bottom-1);
    LineTo(ScreenDC, Left, Top);

    MoveTo(ScreenDC, Left-1, Top-1);
    LineTo(ScreenDC, Right, Top-1);
    LineTo(ScreenDC, Right, Bottom);
    LineTo(ScreenDC, Left-1, Bottom);
    LineTo(ScreenDC, Left-1, Top-1);
  end;

  SetRop2(ScreenDC, OldRop2);
  SelectObject(ScreenDC, OldPen);
  ReleaseDC(0,ScreenDC);
end;



procedure TCalFormDrag.HandleMouseMove(Sender: TObject; Shift: TShiftState; X, Y: Integer);
begin
  { During a solid drag, the form has its dimensions set each time the mouse
    is moved.  During a hollow drag, the previous rectangle must be erased
    (by drawing over it) before drawing the new one }

  with Owner as TForm do
    case DragState of
      fdNone :
        if AllowSize then begin
          if (X > ClientWidth-DragWidth) and (Y > ClientHeight-DragWidth) then begin
            Cursor := crSizeNWSE;
            Exit;
          end
          else begin
            Cursor := crDefault;
          end;
          if Assigned(FMouseMove) then FMouseMove(Sender, Shift, X, Y);
        end;

      fdSolidMove :
        SetBounds(Left+X-xofs, Top+Y-yofs, Width, Height);

      fdSolidSize :
        SetBounds(Left, Top,
          Range(X+xofs, MinWidth, MaxWidth),
          Range(Y+yofs, MinHeight, MaxHeight));

      fdHollowMove:
        begin
          InverseRect;
          FormRect := Bounds(Left+X-xofs, Top+Y-yofs, Width, Height);
          InverseRect;
        end;

      fdHollowSize:
        begin
          InverseRect;
          FormRect := Bounds(Left, Top, Range(X+xofs, MinWidth, MaxWidth),
            Range(Y+yofs, MinHeight, MaxHeight));
          InverseRect;
        end;
    end {case};
end;


procedure TCalFormDrag.HandleMouseDown(Sender: TObject; Button: TMouseButton;
  Shift: TShiftState; X, Y: Integer);
var i: Integer;
begin
  with Owner as TForm do
  if (Button = mbLeft) and not (ssDouble in Shift)
   and (WindowState <> WsMaximized) then begin
    if (X > ClientWidth-DragWidth) and (Y > ClientHeight-DragWidth) and AllowSize then begin
      { Start a resize operation }
      xofs := Width - X;
      yofs := Height - Y;

      if Hollow then begin
        FDragState := fdHollowSize;
        StartHollowDrag;
      end
      else begin
        { Suppress OnPaint and OnResize }
        FDragState := fdSolidSize;
        OnPaint := nil;
        OnResize := nil;
        for i := 0 to ControlCount-1 do Controls[i].Hide;
        Canvas.Brush.Color := Color;
        Canvas.FillRect(Rect(0, 0, ClientWidth, ClientHeight));
      end;
    end
    else if AllowMove then begin
      { Start a move operation }
      xofs := X;
      yofs := Y;
      if Hollow then begin
        FDragState := fdHollowMove;
        StartHollowDrag;
      end
      else FDragState := fdSolidMove;
    end
    else if Assigned(FMouseDown) then
      { trigger form's original OnMouseDown event }
      FMouseDown(Sender, Button, Shift, X, Y);
  end
  else if Assigned(FMouseDown) then
     FMouseDown(Sender, Button, Shift, X, Y);
end;


procedure TCalFormDrag.HandleMouseUp(Sender: TObject; Button: TMouseButton;
  Shift: TShiftState; X, Y: Integer);
var i: Integer;
begin
  if (Button = mbLeft) and (FDragState <> fdNone) then
  with Owner as TForm do begin
    { End the drag operation }
    if FDragState = fdSolidSize then begin
      { Tell the form that it has just been resized, and force it
        to paint if required.  }
      FDragState := fdNone;
      if Assigned(FResize) then FResize(Owner);
      if Assigned(FPaint) then FPaint(Owner);
      OnResize := FResize;
      OnPaint := FPaint;
      for i := 0 to ControlCount-1 do Controls[i].Show;
    end
    else if FDragState in [fdHollowSize, fdHollowMove] then begin
      { Set the new coordinates of the form and clean up the
        graphics stuff }
      ReleaseCapture;
      InverseRect;
      BoundsRect := FormRect;
      DeleteObject(InversePen);
    end;
    FDragState := fdNone;
  end
  else
    if Assigned(FMouseUp) then FMouseUp(Sender, Button, Shift, X, Y);
end;

function Min(a, b: Integer): Integer; assembler;
asm
  MOV	AX, a
  CMP	AX, b
  JLE	@@1
  MOV	AX, b
@@1:
end;


function Max(a, b: Integer): Integer; assembler;
asm
  MOV	AX, a
  CMP	AX, b
  JGE	@@1
  MOV	AX, b
@@1:
end;

function Range(n, lower, upper: Integer): Integer; assembler;
asm
   MOV  AX, n
   CMP  AX, lower
   JGE  @@1
   MOV  AX, lower
   JMP  @finish
@@1:
   CMP  AX, upper
   JLE  @finish
   MOV  AX, upper
   JMP  @finish
@@2:
   MOV  AX, lower
@finish:
end;


function Sign(x: Integer) : Integer; assembler;
asm
   MOV  AX, X
   CMP  AX, 0
   JL   @@1
   JG   @@2
   XOR  AX, AX
   JMP  @finish
@@1:
   MOV  AX, -1
   JMP  @finish
@@2:
   MOV  AX, 1
@finish:
end;



procedure Border3d(Canvas : TCanvas; Width, Height: Integer);
begin
  with Canvas do begin
    Pen.Color := clBtnHighLight;
    MoveTo(0, Height);
    LineTo(0, 0);
    LineTo(Width, 0);
    Pen.Color := clBtnShadow;
    LineTo(Width, Height);
    LineTo(0, Height);
  end;
end;


procedure ErrorMsg(const msg: string);
begin
  MsgDialog(msg, mtError, [mbOK], 0);
end;


procedure PlaySound(const filename: TFilename);
var s: TFilename;
begin
  if CompareText(filename, '<None>') <> 0 then
    SndPlaySound(StrPCopy(@s, filename), SND_ASYNC or SND_NODEFAULT);
end;



function Intersects(const R, S: TRect): Boolean;
var dummy: TRect;
begin
  Result := IntersectRect(dummy, R, S) > 0;
end;

function NormalizeRect(p, q: TPoint): TRect; assembler;
asm
  MOV  AX, p.x
  MOV  BX, p.y
  MOV  CX, q.x
  MOV  DX, q.y
  CMP  AX, CX
  JLE  @@1
  XCHG AX, CX
@@1:
  CMP  BX, DX
  JLE  @@2
  XCHG BX, DX
@@2:
  LES  DI, @Result
  MOV  TRect(ES:[DI]).Left, AX
  MOV  TRect(ES:[DI]).Top, BX
  MOV  TRect(ES:[DI]).Right, CX
  MOV  TRect(ES:[DI]).Bottom, DX
end;



function TimeStampToDate(FileDate: Longint): TDateTime;
begin
  try Result := FileDateToDateTime(FileDate)
  except on EConvertError do Result := Date;
  end;
end;

function GetRegValue(key : string): string;
var cb : Longint;
begin
  cb := 255;
  if RegQueryValue(HKEY_CLASSES_ROOT, StringAsPChar(key),
     @Result[1], cb) = ERROR_SUCCESS then
    Result[0] := Chr(cb-1)
  else
    Result := '';
end;


function GetRadioIndex(const R: array of TRadioButton): Integer;
var i: Integer;
begin
  for i := 0 to High(R) do
    if R[i].Checked then begin
      Result := i;
      exit;
    end;
  Result := 0;
end;


procedure SetRadioIndex(const R: array of TRadioBUtton; index: Integer);
var i: Integer;
begin
  for i := 0 to High(R) do R[i].Checked := i = index;
end;


function GetMenuCheck(const M: array of TMenuItem): Integer;
var i: Integer;
begin
  for i := 0 to High(M) do
    if M[i].Checked then begin
      Result := i;
      exit;
    end;
  Result := 0;
end;


procedure SetMenuCheck(const M: array of TMenuItem; index: Integer);
var i: Integer;
begin
  for i := 0 to High(M) do M[i].Checked := i = index;
end;


procedure RefreshCursor;
var p: TPoint;
begin
  GetCursorPos(p);
  SetCursorPos(p.x, p.y);
end;


function AddHistory(Combo : TComboBox): Boolean;
begin
  Result := False;
  with Combo, Combo.Items do
    if (Text <> '') and (IndexOf(Text) = -1) then begin
      Result := True;
      Insert(0, Text);
      if Count > 24 then Delete(Count-1);
    end;
end;

function MsgDialog(const Msg: string; AType: TMsgDlgType;
  AButtons: TMsgDlgButtons; HelpCtx: Longint): Word;
const
  Sound : array[TMsgDlgType] of Word =
    (MB_ICONEXCLAMATION, MB_ICONHAND, MB_OK, MB_ICONQUESTION, 0);
begin
  if MsgDialogSounds and (AType < mtCustom) then MessageBeep(Sound[AType]);
  Result := MessageDlg(Msg, AType, AButtons, HelpCtx);
end;


function ShowModalDialog(FormClass : TFormClass): TModalResult;
begin
  with FormClass.Create(Application) do
  try
    Result := ShowModal;
  finally
    Free;
  end;
end;


function InitBitmap(ABitmap: TBitmap;
  AWidth, AHeight : Integer; Color : TColor) : TBitmap;
begin
  { initializes a bitmap with width, height and background colour }

  with ABitmap do begin
    Width := AWidth;
    Height := AHeight;
    Canvas.Brush.Color := Color;
    Canvas.FillRect(Rect(0, 0, Width, Height));
  end;
  Result := ABitmap;
end;


procedure ShrinkIcon(H : HIcon; Glyph : TBitmap);
var
  bmp : TBitmap;
  i, j : Integer;
  src, dest : HDC;
begin
  bmp := InitBitmap(TBitmap.Create, 32, 32, clSilver);
  DrawIcon(bmp.Canvas.Handle, 0, 0, H);

  try
    with Glyph do begin
      Width := 16;
      Height := 16;

      Canvas.StretchDraw(Rect(0, 0, 16, 16), bmp);
      src := bmp.Canvas.Handle;
      dest := Canvas.Handle;

      for i := 0 to 15 do
        for j := 0 to 15 do
          if GetPixel(dest, i, j) = clSilver then
           SetPixel(dest, i, j, GetPixel(src, i shl 1, j shl 1));

      Canvas.Pixels[0, 15] := clBtnFace;
    end;
  finally
    bmp.Free;
  end;
end;

constructor TUniqueStrings.Create;
begin
  inherited Create;
  Sorted := True;
  Duplicates := dupIgnore;
end;


function LowCase(c : Char) : Char; assembler;
asm
      MOV    AL, c
      CMP    AL, 'A'
      JB     @Finish
      CMP    AL, 'Z'
      JA     @Finish
      ADD    AL, 32
@Finish:
end;


function InString(c: Char; const s: string): Boolean; assembler;
asm
      XOR     AH,AH
      LES     DI,s
      MOV     AL,ES:[DI]
      INC     DI
      MOV     CH,AH
      MOV     CL,AL
      MOV     AL,c
      CLD
      REPNE   SCASB
      JNE     @@1
      INC     AH
@@1:  MOV     AL,AH
end;


function CharCount(c: Char; const S: string): Integer; assembler;
asm
      XOR     AH,AH
      LES     DI,S
      MOV     AL,ES:[DI]
      INC     DI
      MOV     CX, AX
      MOV     AL,c
      CLD
@@1:  REPNE   SCASB
      JNE     @@2
      INC     AH
      JMP     @@1
@@2:  MOV     AL,AH
      XOR     AH,AH
end;


function Blank(const s: string): boolean; assembler;
asm
      LES     DI, s
      SUB     CX, CX
      MOV     CL, BYTE PTR ES:[DI]
      JCXZ    @@1
      INC     DI
      CLD
      MOV     AL, 32
      REP     SCASB
      JZ      @@1
      MOV     AL, False
      JMP     @@2
@@1:  MOV     AL, True
@@2:
end;


function MakePath(const s: string): string;
begin
  Result := s;
  if Result[Length(Result)] <> '\' then AppendStr(Result, '\');
end;


function MakeDirname(const s: string): string;
begin
  Result := s;
  if (Length(Result) <> 3) and (Result[Length(Result)] = '\') then
    Dec(Result[0]);
end;

function ExtractFileDir(const s: string): string;
begin
  Result := ExtractFilePath(s);
  if (Length(Result) <> 3) and (Result[Length(Result)] = '\') then
    Dec(Result[0]);
end;




function FillString(c: char; n: Byte): string; assembler;
asm
      MOV CL, n
      XOR CH, CH
      LES DI, @result
      MOV BYTE PTR ES:[DI], CL
      INC DI
      MOV AL, c
      CLD
      REP STOSB
end;


function GetStrKey(const s: string): string;
var i: Integer;
begin
  Result := s;
  i := Pos('=', Result);
  if i > 0 then Result[0] := Chr(i-1);
end;


function GetStrValue(const s: string): string;
var i: Integer;
begin
  i := Pos('=', s);
  if i = 0 then Result := '' else Result := Copy(s, i+1, Length(s)-i);
end;


function SetStrValue(const s, value: string): string;
var i: Integer;
begin
  i := Pos('=', s);
  if i = 0 then Result := s + '=' + value
  else Result := Copy(s, 1, i-1) + '=' + value;
end;


function FormatByte(size : Longint): string;
begin
   {
   if size < 1024 then
      if size = 1 then Result := '1 byte'
      else Result := IntToStr(size) + ' bytes'
   else if size < 1048576 then
      Result := FloatToStrF(size / 1024, ffNumber, 7, 2) + 'KB'
   else
      Result := FloatToStrF(size / 1048576, ffNumber, 7, 2) + 'MB';
   }
   if size < 1024 then
      if size = 1 then Result := '1 byte'
      else Result := Format('%d bytes', [size])
   else if size < 1048576 then
      Result := Format('%.2n KB', [size / 1024])
   else
      Result := Format('%.2n MB', [size / 1048576]);
end;


function Unformat(const source, pattern: string; const args: array of const): Integer;
var
  i, j, argindex, start, finish, maxlen: Integer;
  c : Char;
begin
  Result := 0;
  argindex := 0;
  i := 1;
  j := 1;
  while (i < Length(pattern)) and (j <= Length(source)) do begin

    if pattern[i] = '%' then
      case pattern[i+1] of
       'D' : begin
               Inc(i, 2);
               while (j <= Length(source)) and
                 ((source[j] in Digits) or (source[j] = '-')) do Inc(j);
               Inc(Result);
             end;

       'S' : begin
               Inc(i, 2);
               if i > Length(pattern) then break
               else begin
                 c := pattern[i];
                 while (j <= Length(source)) and (source[j] <> c) do
                   Inc(j);
               end;
               Inc(Result);
             end;

       'd' : begin
               if argindex > High(args) then
                 raise EConvertError.Create('Not enough arguments');
               Inc(i, 2);
               start := j;
               while (j <= Length(source)) and
                 ((source[j] in Digits) or (source[j] = '-')) do
                 Inc(j);
               finish := j;
               if finish > start then
                 PInteger(args[argindex].VPointer)^ :=
                   StrToInt(Copy(source, start, finish - start));
               Inc(argindex);
               Inc(Result);
             end;

       's' : begin
               if argindex > High(args)-1 then
                 raise EConvertError.Create('Not enough arguments');

               if args[argindex+1].VType <> vtInteger then
                   raise EConvertError.Create('No string size specified');

               maxlen := args[argindex+1].VInteger;

               Inc(i, 2);
               if i > Length(pattern) then begin
                 args[argindex].VString^ :=
                   Copy(source, j, Min(Length(source) + 1 - j, maxlen));
                 Inc(argindex);
                 break;
               end
               else begin
                 c := pattern[i];
                 start := j;
                 while (j <= Length(source)) and (source[j] <> c) do
                   Inc(j);
                 finish := j;

                 args[argindex].VString^ := Copy(source, start,
                   Min(finish - start, maxlen));
                 Inc(argindex, 2);
               end;
               Inc(Result);
             end;
      else Inc(i);
      end
    else
      if pattern[i] <> source[j] then
        raise EConvertError.Create('Pattern mismatch!')
      else begin
        Inc(i);
        Inc(j);
      end;
  end;
end;


function FileParams(files: TStrings): string;
var
  i: Integer;
begin
  Result := '';
  i := 0;
  while (i < files.Count) and (Length(Result) < 255) do begin
    if Pos('.', files[i]) = 0 then AppendStr(Result, files[i] + '. ')
    else AppendStr(Result, files[i] + ' ');
    Inc(i);
  end;
end;


function GetWord(var s: string): string;
var i: Integer;
begin
  i := Pos(' ', s);
  if i = 0 then begin
    if Length(s) > 0 then begin
      Result := s;
      s := '';
    end
    else Result := '';
  end
  else begin
    Result := Copy(s, 1, i-1);
    while (i <= Length(s)) and (s[i] = ' ') do Inc(i);
    Delete(s, 1, i-1);
  end;
end;


function StringAsPChar(var s: OpenString): PChar;
begin
  Result := @s[1];
  if Length(s) = High(s) then Dec(s[0]);
  s[Length(s) + 1] := #0;
end;


function AttrToStr(attr : Integer): TAttrStr; assembler;
asm
     LES  DI, @Result
     MOV  BX, DI
     XOR  AX, AX
     INC  DI
     MOV  CX, attr

     MOV  DX, CX
     AND  DX, faArchive
     JZ   @@1
     MOV  AL, 'a'
     STOSB
     INC  AH

@@1: MOV  DX, CX
     AND  DX, faReadOnly
     JZ   @@2
     MOV  AL, 'r'
     STOSB
     INC  AH

@@2: MOV  DX, CX
     AND  DX, faHidden
     JZ   @@3
     MOV  AL, 'h'
     STOSB
     INC  AH

@@3: MOV  DX, CX
     AND  DX, faSysfile
     JZ   @@4
     MOV  AL, 's'
     STOSB
     INC  AH

@@4: MOV  DX, CX
     AND  DX, faDirectory
     JZ   @@5
     MOV  AL, 'd'
     STOSB
     INC  AH
@@5:
     MOV  BYTE PTR ES:[BX], AH
end;


function LTrim(const s: string): string;
var
  i: Integer;
begin
  i := 1;
  while (i <= Length(s)) and (s[i] = ' ') do Inc(i);
  Result := Copy(s, i, 255);
end;


function RTrim(const s: string): string;
var
  i: Integer;
begin
  i := Length(s);
  while (s[i] = ' ') do Dec(i);
  Result := Copy(s, 1, i);
end;


function Trim(const s: string): string;
begin
  Result := LTrim(RTrim(s))
end;


{------------------------------------------------------------------------------------}
procedure Register;
begin
  RegisterComponents('CalNew', [TCalIconDialog]);
  RegisterComponents('CalNew', [TCalAppHolder]);
  RegisterComponents('CalNew', [TCalBarGauge]);
  RegisterComponents('CalNew', [TCalStyleSpeed]);
  RegisterComponents('CalNew', [TCalCheckList]);
  RegisterComponents('CalNew', [TCalMultiGrid]);
  RegisterComponents('CalNew', [TCalDropClient]);
  RegisterComponents('CalNew', [TCalDropServer]);
  RegisterComponents('CalNew', [TCalSystemMenu]);
  RegisterComponents('CalNew', [TCalFormDrag]);
end;

procedure Done; far;
begin
  Clients.Free;
end;

initialization
  AddExitProc(DoneCheckList);
  Clients := TList.Create;
  AddExitProc(Done);
  ApplicationPath := ExtractFilePath(ParamStr(0));
  WinPath[0] := Chr(GetWindowsDirectory(@WinPath[1], 79));
  WinPath := MakePath(WinPath);
end.
